! ===========================================================================================================
! module ncprof_arrays.f90
!
! -----------------------------------------------------------------------------------------------------------
!
! Revision history
! date     who    what
! 20190301 JN     Handle the files with no primary core argo profile, but only secondary profile (from SIO & PMEL)
! 20190212 JN     Fix the discrepancy between units and values in the QC/NCPROF files for the floats containing 
!                 CONDUCTIVITY.
! 20180619 VH     Do NOT pur interpolated temp/psal into the ncprof R file
! 20180417 VH	  Fix slight prblem with PROFILE_CNDC_QC. (installed, not documented for Mar 21 qcplus.x
!                 General WARNING revisions.
! 20180303 VH	  Use the PSAL_QC for DNCD_QC, also PROFILE_PSAL_QC for PROFILE_CNDC_QC
! 20180307 VH     "no primary profile" changes merged into the new BR version
! 20180111 JN     "no primary profile": inplement nprf_start when creating ncprof (VH merged 20180307)
! 20170712 all    Release of qcplus v1.1 -   mmany changes to all, some documented/tracked, some not
! 20160511 all    Initial release of qcplus v1.1.0
!
! ===========================================================================================================
! Notes:

! This module contains variable/array definitions used by write_ncprof to create the Core-ARGO profile nc file.
! ===========================================================================================================

module ncprof_arrays

use communication                                                ! module
use data_arrays                                                  ! module
use netcdf_arrays                                                ! module

implicit none

! ===========================================================================================================
! moedule varialbe/array declarations
! ===========================================================================================================

integer :: PROJECT_NAME_id,           PROJECT_NAME_dims(NDIMS2);          character(len=64) :: PROJECT_NAME(MAXPRF);
integer :: PI_NAME_id,                PI_NAME_dims(NDIMS2);               character(len=64) :: PI_NAME(MAXPRF);
integer :: STATION_PARS_id      ,     STATION_PARAMETERS_dims(ND3);       character(len=16) :: STATION_PARAMETERS(MAXPAR);
! note on simple variable, not 2-D array for STATION_PARAMETERS: All profiles have the same parameters, so just deinfe one array,
! then load it once for each profile

integer :: CYCLE_NUMBER_id,           CYCLE_NUMBER_dims(NDIMS1);          integer :: CYCLE_NUMBER(MAXPRF)=9999;
integer :: DIRECTION_id,              DIRECTION_dims(NDIMS1);             character(len=1) :: DIRECTION(MAXPRF)=' ';
integer :: DC_REFERENCE_id,           DC_REFERENCE_dims(NDIMS2);          character(len=32) :: DC_REFERENCE(MAXPRF)=' ';
integer :: DATA_STATE_IND_id,         DATA_STATE_IND_dims(NDIMS2);        character(len=4) :: DATA_STATE_INDICATOR(MAXPRF)='    ';
character(len=49) :: DATA_STATE_IND_long_name = 'Degree of processing the data have passed through'
integer :: DATA_MODE_id,              DATA_MODE_dims(NDIMS1);               character(len=1) :: DATA_MODE(MAXPRF);
character(len=62) :: DATA_MODE_conventions = 'R : real time; D : delayed mode; A : real time with adjustment'
integer :: PLATFORM_TYPE_id,          PLATFORM_TYPE_dims(NDIMS2);           character(len=32) :: PLATFORM_TYPE(MAXPRF);

integer :: FLOAT_SERIAL_NO_id,        FLOAT_SERIAL_NO_dims(NDIMS2);       character(len=32) :: FLOAT_SERIAL_NO(MAXPRF);
                                                                          character(len=16) :: FLOAT_SERIAL_NO_3p0(MAXPRF);
integer :: FIRMWARE_VERSION_id,       FIRMWARE_VERSION_dims(NDIMS2);      character(len=32) :: FIRMWARE_VERSION(MAXPRF);
                                                                          character(len=16) :: FIRMWARE_VERSION_3p0(MAXPRF);

integer :: WMO_INST_TYPE_id,          WMO_INST_TYPE_dims(NDIMS2);         character(len=4) :: WMO_INST_TYPE(MAXPRF)=" ";
integer :: JULD_id,                   JULD_dims(NDIMS1);                  double precision :: JULD(MAXPRF) = FILLDBLE;
integer :: JULD_QC_id,                JULD_QC_dims(NDIMS1);               character(len=1) :: JULD_QC(MAXPRF)=' ';
integer :: JULD_LOCATION_id,          JULD_LOCATION_dims(NDIMS1);         double precision :: JULD_LOCATION(MAXPRF) = FILLDBLE;
integer :: POSITION_QC_id,            POSITION_QC_dims(NDIMS1);           character(len=1) :: POSITION_QC(MAXPRF)=' ';
integer :: LATITUDE_id,               LATITUDE_dims(NDIMS1);              double precision :: LATITUDE(MAXPRF)= FILLPOSN;
integer :: LONGITUDE_id,              LONGITUDE_dims(NDIMS1);             double precision :: LONGITUDE(MAXPRF)= FILLPOSN;
integer :: POSITIONING_SYSTEM_id,     POSITIONING_SYSTEM_dims(NDIMS2);    character(len=8) :: POSITIONING_SYSTEM(MAXPRF)=" ";;
integer :: VERT_SAMPLING_SCHEME_id,   VERT_SAMPLING_SCHEME_dims(NDIMS2);  character(len=256) :: VERT_SAMPLING_SCHEME(MAXPRF)=" ";
integer :: CONFIG_MISSION_NBR_id,     CONFIG_MISSION_NBR_dims(NDIMS1);    integer :: CONFIG_MISSION_NUMBER(MAXPRF)=FILLINT;
!================================================================================
! <PARAM> id definitions
!================================================================================
integer :: PRES_ids(5), TEMP_ids(5), PSAL_ids(5), DOXY_ids(5), CNDC_ids(5)
integer :: PROF_QC_ids(MAXPAR);		! nc ids for Profile_<>)qc
integer :: PROF_PRES_ID,PROF_TEMP_ID,PROF_PSAL_ID,PROF_DOXY_ID,PROF_CNDC_ID
!================================================================================
! 20Jun2016 - separate ncprof & nctraj defs for valic_min,valid_max and long_name ncprof=v3.03, nctraj=v3.1
! Valid_min/max                               PRES     TEMP     CNDC     PSAL    DOXY
    real,parameter :: PAR_vmin(MAXPAR) = (/    0.0,    -2.5,     0.0,     2.0,  0.0    /);  ! User Man 3.03
    real,parameter :: PAR_vmax(MAXPAR) = (/12000.0,    40.0,     8.5,    41.0,   650.0  /); ! User Man 3.03

! Valid_min/max                                   PRES     TEMP     CNDC     PSAL    DOXY
    real,parameter :: PAR_vmin_3p0(MAXPAR) = (/    0.0,    -2.0,     0.0,     0.0,  0.0    /);	! User Man 3.03
    real,parameter :: PAR_vmax_3p0(MAXPAR) = (/12000.0,    40.0,     8.5,    42.0,   650.0  /);	! User Man 3.03

    character(len=86) :: PAR_lnam(MAXPAR) =  (/ & ! User Man 3.03 for VARIABLE & ADJUSTED &  ERROR fields profver3p0
	'SEA PRESSURE                                                                          ',  & ! PRES profver3p0
	'SEA TEMPERATURE IN SITU ITS-90 SCALE                                                  ',  & ! TEMP profver3p0
	'ELECTRICAL CONDUCTIVITY                                                               ',  & ! CNDC profver3p0
	'PRACTICAL SALINITY                                                                    ',  & ! PSAL profver3p0
	'DISSOLVED OXYGEN                                                                      ' /) ! DOXY profver3p0
    character(len=86) :: PAR_lnamV(MAXPAR) =  (/ & ! User Man 3.2 long_name for VARIABLE fields
        'Sea water pressure, equals 0 at sea-level                                             ',  & ! PRES v3p1
        'Sea temperature in-situ ITS-90 scale                                                  ',  & ! TEMP v3p1
        'Electrical conductivity                                                               ',  & ! CNDC v3p1
        'Practical salinity                                                                    ',  & ! PSAL v3p1
        'Dissolved oxygen                                                                      ' /)  ! DOXY v3p1
    character(len=86) :: PAR_lnamA(MAXPAR) =  (/ & ! User Man 3.2 long_name for ADJUSTED fields
        'Sea water pressure, equals 0 at sea-level                                             ',  & ! PRES v3p1
        'Sea temperature in-situ ITS-90 scale                                                  ',  & ! TEMP v3p1
        'Electrical conductivity                                                               ',  & ! CNDC v3p1
        'Practical salinity                                                                    ',  & ! PSAL v3p1
        'Dissolved oxygen                                                                      ' /)  ! DOXY v3p1
    character(len=86) :: PAR_lnamE(MAXPAR) =  (/ & ! User Man 3.2 long_name for ERROR fields v3p1
        'Contains the error on the adjusted values as determined by the delayed mode QC process',  & ! PRES v3p1
        'Contains the error on the adjusted values as determined by the delayed mode QC process',  & ! TEMP v3p1
        'Contains the error on the adjusted values as determined by the delayed mode QC process',  & ! CNDC v3p1
        'Contains the error on the adjusted values as determined by the delayed mode QC process',  & ! PSAL v3p1
        'Dissolved oxygen                                                                      ' /)               ! DOXY v3p1

!================================================================================
! resolution and C/FORTRAN_format attributes - moved to netcdf_arrays
!================================================================================
!   ! SCIENTIFIC_CALIB declarations
!================================================================================
integer,parameter :: N_CALIB = 1; public N_CALIB
integer :: SCICAL_ids(5), SCICAL_dims(ND4);
character(len=16) :: PARAMETER
character(len=256) :: SCICAL_eqtn
character(len=256) :: SCICAL_coef
character(len=256) :: SCICAL_cmnt
character(len=14)  :: SCICAL_date

!================================================================================
! HISTORY declarations
!================================================================================
integer,parameter :: N_HISTORY = NF_UNLIMITED; public N_HISTORY
integer,parameter :: N_HISTMAX = 4; ! 2 regular, plus 1 each for adj_pres,adj_psal
integer :: HIST_ids(12)
integer :: HIST_dim2(ND2), HIST_dim3(ND3)			! See later section explaning why we need dim2 & dim3

character(len=4)  :: NHIST_inst;   !  1 HIST_INSTITUTION
character(len=4)  :: NHIST_step;   !  2 HIST_STEP
character(len=4)  :: NHIST_soft;   !  3 HIST_SOFTWARE
character(len=4)  :: NHIST_sfrl;   !  4 NHIST_SOFTWARE_RELEASE
character(len=64) :: NHIST_refr;   !  5 NHIST_REFERENCE
character(len=14) :: NHIST_date;   !  6 NHIST_DATE
character(len=4)  :: NHIST_actn;   !  7 NHIST_ACTION
character(len=16) :: NHIST_parm;   !  8 NHIST_PARAMETER
real              :: NHIST_strt;   !  9 NHIST_START_PRES
real              :: NHIST_stop;   ! 10 NHIST_STOP_PRES
real              :: NHIST_prev;   ! 11 NHIST_PREVIOUS_VALUE
character(len=16) :: NHIST_qctest; ! 12 NHIST_QCTEST

! ===========================================================================================================

contains

! ===========================================================================================================
! module functions/subroutines
! ===========================================================================================================

subroutine def_ncprof_vars(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;
    integer :: loc = 0;

    character(len=15), parameter :: sub = 'def_ncprof_vars'
    double precision, parameter :: minlat =  -90.0;
    double precision, parameter :: maxlat =   90.0;
    double precision, parameter :: minlon = -180.0;
    double precision, parameter :: maxlon =  180.0;
    double precision :: doublefill(1);			! Set differently for different variables


    !================================================================================
    PROJECT_NAME_dims(2) = N_PROF_dimid;
    PROJECT_NAME_dims(1) = STRING64_dimid;
    ret=nf_def_var(ncid, 'PROJECT_NAME', NF_CHAR,NDIMS2, PROJECT_NAME_dims,PROJECT_NAME_id); call cke(sub,ret, 5);
    ret=nf_put_att_text(ncid, PROJECT_NAME_id, 'long_name', 19, 'Name of the project'); call cke(sub,ret, 6);
    ret=nf_put_att_text(ncid, PROJECT_NAME_id, '_FillValue', 1, ' '); call cke(sub,ret, 7);
    !==================================================================
    PI_NAME_dims(2) = N_PROF_dimid;
    PI_NAME_dims(1) = STRING64_dimid;
    ret=nf_def_var(ncid, 'PI_NAME', NF_CHAR,NDIMS2, PI_NAME_dims,PI_NAME_id); call cke(sub,ret, 8);
    ret=nf_put_att_text(ncid, PI_NAME_id, 'long_name', 34, 'Name of the principal investigator'); call cke(sub,ret, 9);
    ret=nf_put_att_text(ncid, PI_NAME_id, '_FillValue', 1, ' '); call cke(sub,ret,10);
    !==================================================================
    STATION_PARAMETERS_dims(3) = N_PROF_dimid;		! leftmost dimension		don't ask
    STATION_PARAMETERS_dims(2) = N_PARAM_dimid;		! middle dimensin		 it just works
    STATION_PARAMETERS_dims(1) = STRING16_dimid;	! rightmost dimensin		 ! ncprof PARAMETER names are 16 for coreArgo
    ret=nf_def_var(ncid,'STATION_PARAMETERS',NF_CHAR,NDIMS3,STATION_PARAMETERS_dims,STATION_PARS_id); call cke(sub,ret,33);
    ret=nf_put_att_text(ncid,STATION_PARS_id,'long_name',44,'List of available parameters for the station'); call cke(sub,ret,34);
    ret=nf_put_att_text(ncid,STATION_PARS_id,'conventions',22,'Argo reference table 3'); call cke(sub,ret,35);
    ret=nf_put_att_text(ncid,STATION_PARS_id,'_FillValue',1,' '); call cke(sub,ret,35);
    !==================================================================
    CYCLE_NUMBER_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'CYCLE_NUMBER', NF_INT,NDIMS1, CYCLE_NUMBER_dims,CYCLE_NUMBER_id); call cke(sub,ret,11);
    ret=nf_put_att_text(ncid,CYCLE_NUMBER_id, 'long_name', 18, 'Float cycle number')
    str = '0...N, 0 : launch cycle (if exists), 1 : first complete cycle';
    if(profver3p0) str = '0..N, 0 : launch cycle (if exists), 1 : first complete cycle';
    ret=nf_put_att_text(ncid,CYCLE_NUMBER_id, 'conventions',lt(str),trim(str));
    if(DEBUG)then; write(msg,'("CYCNUM write cycum to prof: ",a)')trim(str); call wrtdbg(sub,msg); endif
    ret=nf_put_att_int(ncid,CYCLE_NUMBER_id, '_FillValue', NF_INT, 1, FILLINT) ! FILLINT int fill value in netcdf_arrays
    !==================================================================
    DIRECTION_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'DIRECTION', NF_CHAR,NDIMS1, DIRECTION_dims,DIRECTION_id); call cke(sub,ret,12);
    ret=nf_put_att_text(ncid,DIRECTION_id,'long_name', 33, 'Direction of the station profiles'); call cke(sub,ret,13);
    ret=nf_put_att_text(ncid,DIRECTION_id,'conventions',45,'A: ascending profiles, D: descending profiles'); call cke(sub,ret,14);
    ret=nf_put_att_text(ncid,DIRECTION_id,'_FillValue', 1, ' '); call cke(sub,ret,15);
    !================================================================================
    DATA_CENTRE_dims(2) = N_PROF_dimid ;
    DATA_CENTRE_dims(1) = STRING2_dimid;
    ret=nf_def_var(ncid, 'DATA_CENTRE', NF_CHAR,NDIMS2, DATA_CENTRE_dims,DATA_CENTRE_id); call cke(sub,ret,16);  ! For ncprof
    ret=nf_put_att_text(ncid, DATA_CENTRE_id, 'long_name', 46, DC_long_name_prof);          call cke(sub,ret,17)
    ret=nf_put_att_text(ncid, DATA_CENTRE_id, 'conventions', 22, 'Argo reference table 4'); call cke(sub,ret,18);
    ret=nf_put_att_text(ncid, DATA_CENTRE_id, '_FillValue', 1, ' '); call cke(sub,ret,19);
    !================================================================================
    DC_REFERENCE_dims(2) = N_PROF_dimid ;
    DC_REFERENCE_dims(1) = STRING32_dimid;
    ret=nf_def_var(ncid, 'DC_REFERENCE', NF_CHAR,NDIMS2, DC_REFERENCE_dims,DC_REFERENCE_id); call cke(sub,ret,20);
    ret=nf_put_att_text(ncid, DC_REFERENCE_id, 'long_name', 40, 'Station unique identifier in data centre'); call cke(sub,ret,21);
    ret=nf_put_att_text(ncid, DC_REFERENCE_id, 'conventions', 22, 'Data centre convention'); call cke(sub,ret,22);
    ret=nf_put_att_text(ncid, DC_REFERENCE_id, '_FillValue', 1, ' '); call cke(sub,ret,23);
    !================================================================================
    DATA_STATE_IND_dims(2) = N_PROF_dimid ;
    DATA_STATE_IND_dims(1) = STRING4_dimid;
    ret=nf_def_var(ncid, 'DATA_STATE_INDICATOR', NF_CHAR,NDIMS2, &
		DATA_STATE_IND_dims,DATA_STATE_IND_id); call cke(sub,ret,24);
    ret=nf_put_att_text(ncid, DATA_STATE_IND_id, 'long_name', 49, DATA_STATE_IND_long_name); call cke(sub,ret,25);
    ret=nf_put_att_text(ncid, DATA_STATE_IND_id, 'conventions', 22, 'Argo reference table 6'); call cke(sub,ret,26);
    ret=nf_put_att_text(ncid, DATA_STATE_IND_id, '_FillValue', 1, ' '); call cke(sub,ret,27);
    !================================================================================
    DATA_MODE_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'DATA_MODE', NF_CHAR,NDIMS1, DATA_MODE_dims,DATA_MODE_id); call cke(sub,ret,28);
    ret=nf_put_att_text(ncid, DATA_MODE_id, 'long_name', 30, 'Delayed mode or real time data'); call cke(sub,ret,29);
    ret=nf_put_att_text(ncid, DATA_MODE_id, 'conventions', 62, DATA_MODE_conventions); call cke(sub,ret,30);
    ret=nf_put_att_text(ncid, DATA_MODE_id, '_FillValue', 1, ' '); call cke(sub,ret,31);
    !==================================================================
    PLATFORM_TYPE_dims(2) = N_PROF_dimid ;
    PLATFORM_TYPE_dims(1) = STRING32_dimid;
    ret=nf_def_var(ncid, 'PLATFORM_TYPE', NF_CHAR,NDIMS2, PLATFORM_TYPE_dims,PLATFORM_TYPE_id); call cke(sub,ret,42);
    ret=nf_put_att_text(ncid, PLATFORM_TYPE_id, 'long_name', 13, 'Type of float'); call cke(sub,ret,43);
    if(.not. profver3p0)then ! NOTE NO OFFSET FOR if(...
    ret=nf_put_att_text(ncid, PLATFORM_TYPE_id, 'conventions',23, 'Argo reference table 23');
    endif ! profver3p0
    ret=nf_put_att_text(ncid, PLATFORM_TYPE_id, '_FillValue', 1, ' '); call cke(sub,ret,44);
    !==================================================================
    FLOAT_SERIAL_NO_dims(2) = N_PROF_dimid ;
    FLOAT_SERIAL_NO_dims(1) = STRING32_dimid;
    if(profver3p0) FLOAT_SERIAL_NO_dims(1) = STRING16_dimid;
    ret=nf_def_var(ncid, 'FLOAT_SERIAL_NO', NF_CHAR,NDIMS2, FLOAT_SERIAL_NO_dims,FLOAT_SERIAL_NO_id); call cke(sub,ret,23);
    ret=nf_put_att_text(ncid, FLOAT_SERIAL_NO_id, 'long_name', 26, 'Serial number of the float'); call cke(sub,ret,24);
    ret=nf_put_att_text(ncid, FLOAT_SERIAL_NO_id, '_FillValue', 1, ' '); call cke(sub,ret,25);
    !==================================================================
    FIRMWARE_VERSION_dims(2) = N_PROF_dimid ;
    FIRMWARE_VERSION_dims(1) = STRING32_dimid;
    if(profver3p0) FIRMWARE_VERSION_dims(1) = STRING16_dimid;
    ret=nf_def_var(ncid, 'FIRMWARE_VERSION', NF_CHAR,NDIMS2, FIRMWARE_VERSION_dims,FIRMWARE_VERSION_id); call cke(sub,ret,54);
    str = 'Instrument firmware version'
    if(profver3p0) str = 'Instrument version';
    ret=nf_put_att_text(ncid, FIRMWARE_VERSION_id, 'long_name',lt(str),trim(str));
    ret=nf_put_att_text(ncid, FIRMWARE_VERSION_id, '_FillValue', 1, ' '); call cke(sub,ret,52);
    !================================================================================
    WMO_INST_TYPE_dims(2) = N_PROF_dimid ;
    WMO_INST_TYPE_dims(1) = STRING4_dimid;
    ret=nf_def_var(ncid, 'WMO_INST_TYPE', NF_CHAR,NDIMS2, WMO_INST_TYPE_dims,WMO_INST_TYPE_id); call cke(sub,ret,32);
    ret=nf_put_att_text(ncid, WMO_INST_TYPE_id, 'long_name', 21, 'Coded instrument type'); call cke(sub,ret,33);
    ret=nf_put_att_text(ncid, WMO_INST_TYPE_id, 'conventions', 22, 'Argo reference table 8'); call cke(sub,ret,34);
    ret=nf_put_att_text(ncid, WMO_INST_TYPE_id, '_FillValue', 1, ' '); call cke(sub,ret,35);
    !================================================================================
    JULD_dims(1) = N_PROF_dimid
    ret=nf_def_var(ncid, 'JULD', NF_DOUBLE,NDIMS1, JULD_dims,JULD_id); call cke(sub,ret,36);
    ret=nf_put_att_text(ncid, JULD_id, 'long_name', 63, &
                'Julian day (UTC) of the station relative to REFERENCE_DATE_TIME'); call cke(sub,ret,37);
    ret=nf_put_att_text(ncid, JULD_id, 'standard_name', 4, 'time'); call cke(sub,ret,37);
    ret=nf_put_att_text(ncid, JULD_id, 'units', 34,'days since 1950-01-01 00:00:00 UTC'); call cke(sub,ret,38);
    ret=nf_put_att_text(ncid, JULD_id, 'conventions', 56,&
	'Relative julian days with decimal part (as parts of day)'); call cke(sub,ret,39);
    if(.not. profver3p0)then ! Note no offset
    ret=nf_put_att_double(ncid, JULD_id, 'resolution', NF_DOUBLE, 1, JULD_RESOLUTION); call cke(sub,ret,39);
    endif; ! profv3p0
    doublefill(1)=999999.; ret=nf_put_att_double(ncid, JULD_id, '_FillValue', NF_DOUBLE,1,doublefill(1)); call cke(sub,ret,40);
    ret=nf_put_att_text(ncid, JULD_id, 'axis', 1, 'T'); call cke(sub,ret,40);
    !================================================================================
    JULD_QC_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'JULD_QC', NF_CHAR,NDIMS1, JULD_QC_dims,JULD_QC_id); call cke(sub,ret,41);
    str = "Quality on date and time";
    if(profver3p0) str = "Quality on Date and Time";
    ret=nf_put_att_text(ncid, JULD_QC_id, 'long_name', lt(str),trim(str)); call cke(sub,ret,42);
    ! OUT if(profver3p0) ret=nf_put_att_text(ncid, JULD_QC_id, 'long_name', 24, 'Quality on Date and Time'); call cke(sub,ret,42);
    ret=nf_put_att_text(ncid, JULD_QC_id, 'conventions', 22, 'Argo reference table 2'); call cke(sub,ret,43);
    ret=nf_put_att_text(ncid, JULD_QC_id, '_FillValue', 1, ' '); call cke(sub,ret,44);
    !================================================================================
    JULD_LOCATION_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'JULD_LOCATION', NF_DOUBLE,NDIMS1, JULD_LOCATION_dims,JULD_LOCATION_id); call cke(sub,ret,45);
    ret=nf_put_att_text(ncid, JULD_LOCATION_id, 'long_name', 64, &
		'Julian day (UTC) of the location relative to REFERENCE_DATE_TIME'); call cke(sub,ret,46);
    ret=nf_put_att_text(ncid, JULD_LOCATION_id, 'units', 34,'days since 1950-01-01 00:00:00 UTC'); call cke(sub,ret,47);
    ret=nf_put_att_text(ncid, JULD_LOCATION_id, 'conventions', 56, &
	'Relative julian days with decimal part (as parts of day)'); call cke(sub,ret,48);
    if(.not. profver3p0)then ! Note no offset
    ret=nf_put_att_double(ncid, JULD_LOCATION_id, 'resolution', NF_DOUBLE, 1, JULD_RESOLUTION); call cke(sub,ret,39);
    endif; ! profv3p0
    doublefill(1)=999999.;
    ret=nf_put_att_double(ncid, JULD_LOCATION_id, '_FillValue', NF_DOUBLE,1,doublefill(1)); call cke(sub,ret,49);
    !================================================================================
    LATITUDE_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'LATITUDE', NF_DOUBLE,NDIMS1, LATITUDE_dims,LATITUDE_id); call cke(sub,ret,50);
    ret=nf_put_att_text(ncid, LATITUDE_id, 'long_name', 38,'Latitude of the station, best estimate'); call cke(sub,ret,51);
    ret=nf_put_att_text(ncid, LATITUDE_id, 'standard_name', 8,'latitude'); call cke(sub,ret,52);
    ret=nf_put_att_text(ncid, LATITUDE_id, 'units', 12, 'degree_north'); call cke(sub,ret,53);
    doublefill(1)=99999. ;
    ret=nf_put_att_double(ncid, LATITUDE_id, '_FillValue', NF_DOUBLE,1,doublefill(1)); call cke(sub,ret,54);
    ret=nf_put_att_double(ncid, LATITUDE_id, 'valid_min', NF_DOUBLE, 1, minlat); call cke(sub,ret,55);
    ret=nf_put_att_double(ncid, LATITUDE_id, 'valid_max', NF_DOUBLE, 1, maxlat); call cke(sub,ret,56);
    ret=nf_put_att_text(ncid, LATITUDE_id, 'axis', 1, 'Y'); call cke(sub,ret,57);
    !================================================================================
    LONGITUDE_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'LONGITUDE', NF_DOUBLE,NDIMS1, LONGITUDE_dims,LONGITUDE_id); call cke(sub,ret,58);
    ret=nf_put_att_text(ncid, LONGITUDE_id, 'long_name', 39,'Longitude of the station, best estimate'); call cke(sub,ret,59);
    ret=nf_put_att_text(ncid, LONGITUDE_id, 'standard_name', 9,'longitude'); call cke(sub,ret,60);
    ret=nf_put_att_text(ncid, LONGITUDE_id, 'units', 11, 'degree_east'); call cke(sub,ret,61);
    doublefill(1)=99999. ;
    ret=nf_put_att_double(ncid, LONGITUDE_id, '_FillValue', NF_DOUBLE,1,doublefill(1)); call cke(sub,ret,62);
    ret=nf_put_att_double(ncid, LONGITUDE_id, 'valid_min', NF_DOUBLE, 1, minlon); call cke(sub,ret,63);
    ret=nf_put_att_double(ncid, LONGITUDE_id, 'valid_max', NF_DOUBLE, 1, maxlon); call cke(sub,ret,64);
    ret=nf_put_att_text(ncid, LONGITUDE_id, 'axis', 1, 'X'); call cke(sub,ret,65);
    !================================================================================
    POSITION_QC_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'POSITION_QC', NF_CHAR,NDIMS1, POSITION_QC_dims,POSITION_QC_id); call cke(sub,ret,66);
    ret=nf_put_att_text(ncid,POSITION_QC_id,'long_name',44,'Quality on position (latitude and longitude)'); call cke(sub,ret,67);
    ret=nf_put_att_text(ncid, POSITION_QC_id, 'conventions', 22, 'Argo reference table 2'); call cke(sub,ret,68);
    ret=nf_put_att_text(ncid, POSITION_QC_id, '_FillValue', 1, ' '); call cke(sub,ret,69);
    !================================================================================
    POSITIONING_SYSTEM_dims(2) = N_PROF_dimid ;
    POSITIONING_SYSTEM_dims(1) = STRING8_dimid;
    ret=nf_def_var(ncid,'POSITIONING_SYSTEM',NF_CHAR,NDIMS2,POSITIONING_SYSTEM_dims,POSITIONING_SYSTEM_id); call cke(sub,ret,70);
    ret=nf_put_att_text(ncid, POSITIONING_SYSTEM_id, 'long_name', 18, 'Positioning system'); call cke(sub,ret,71);
    ret=nf_put_att_text(ncid, POSITIONING_SYSTEM_id, '_FillValue', 1, ' '); call cke(sub,ret,72);
    !==================================================================
    VERT_SAMPLING_SCHEME_dims(2) = N_PROF_dimid ;
    VERT_SAMPLING_SCHEME_dims(1) = STRING256_dimid;
    ret=nf_def_var(ncid, 'VERTICAL_SAMPLING_SCHEME', NF_CHAR,NDIMS2, VERT_SAMPLING_SCHEME_dims, &
				VERT_SAMPLING_SCHEME_id); call cke(sub,ret,61);
    ret=nf_put_att_text(ncid, VERT_SAMPLING_SCHEME_id, 'long_name', 24, 'Vertical sampling scheme'); call cke(sub,ret,62);
    ret=nf_put_att_text(ncid, VERT_SAMPLING_SCHEME_id, 'conventions', 23, 'Argo reference table 16');
    ret=nf_put_att_text(ncid, VERT_SAMPLING_SCHEME_id, '_FillValue', 1, ' '); call cke(sub,ret,63);
    !================================================================================
    CONFIG_MISSION_NBR_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid,'CONFIG_MISSION_NUMBER',NF_INT,NDIMS1,CONFIG_MISSION_NBR_dims,CONFIG_MISSION_NBR_id); call cke(sub,ret,11);
    str = "Unique number denoting the missions performed by the float";
    if(profver3p0) str = "Float's mission number for each profile";;
    ret=nf_put_att_text(ncid,CONFIG_MISSION_NBR_id, 'long_name',lt(str),trim(str));
    str = "1...N, 1 : first complete mission";
    if(profver3p0) str = "0..N, 0 : launch mission (if exists), 1 : first complete mission";
    ret=nf_put_att_text(ncid,CONFIG_MISSION_NBR_id,'conventions',lt(str),trim(str));
    ret=nf_put_att_int(ncid,CONFIG_MISSION_NBR_id, '_FillValue', NF_INT, 1, FILLINT) ! FILLINT int fill value in netcdf_arrays
    !==================================================================

end subroutine def_ncprof_vars

  !====================================================================
subroutine put_ncprof_vars(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;
    integer :: loc = 0;
    integer :: is;			! Use to count the station parameters
    integer :: nsp;			! The total number of STATION_PARAMETERS
    integer :: ihst, ipar, ical ! Counters for the dimensions

    integer :: start3(3), count3(3);	! corners for call for STATION_PARAMETERS

    character(len=4) :: c4		! Array messge transfer

    character(len=15), parameter :: sub = 'put_ncprof_vars'

    !================================================================================
    do i=1,N_PROF; PROJECT_NAME(i) = chr_PROJECT_NAME; enddo
     ret=nf_put_var_text(ncid, PROJECT_NAME_id, PROJECT_NAME); call cke(sub,ret,74);
    !================================================================================
    do i=1,N_PROF; PI_NAME(i) = var_PI_NAME; enddo
     ret=nf_put_var_text(ncid, PI_NAME_id, PI_NAME); call cke(sub,ret,75);
    !================================================================================
    is = 0; ! Count the station parameters
    if(hasPRES(0))then;                is=is+1; STATION_PARAMETERS(is) = 'PRES            '; ENDIF
    if(hasTEMP(0))then;                is=is+1; STATION_PARAMETERS(is) = 'TEMP            '; endif
    if(hasCNDC(0))then;                is=is+1; STATION_PARAMETERS(is) = 'CNDC            '; endif
    if(hasPSAL(0))then;                is=is+1; STATION_PARAMETERS(is) = 'PSAL            '; endif
    if(hasDOXY(0) .and. outRDOXY)then; is=is+1; STATION_PARAMETERS(is) = 'DOXY            '; endif
    nsp = is;

    do iprf = 1,N_PROF
	start3(1) = 1;  start3(2) = 1;      start3(3) = iprf;
	count3(1) = 16; count3(2) = nparam; count3(3) = 1;
	 ret=nf_put_vara_text(ncid, STATION_PARS_id,start3, count3,STATION_PARAMETERS); call cke(sub,ret,36);
    enddo


    do i=1,N_PROF; CYCLE_NUMBER(i) = thiscyc; enddo;
     ret=nf_put_var_int(ncid, CYCLE_NUMBER_id, CYCLE_NUMBER); call cke(sub,ret,76);
    !================================================================================
    do i=1,N_PROF; DIRECTION(i) = chr_DIRECTION; enddo
     ret=nf_put_var_text(ncid, DIRECTION_id, DIRECTION); call cke(sub,ret,77);
    !================================================================================
    do i=1,N_PROF; DATA_CENTRE(i) = chr_DATA_CENTRE; enddo
     ret=nf_put_var_text(ncid, DATA_CENTRE_id, DATA_CENTRE); call cke(sub,ret,78);
    !================================================================================
    do i=1,N_PROF; DC_REFERENCE(i) = chr_DC_REFERENCE; enddo
     ret=nf_put_var_text(ncid, DC_REFERENCE_id, DC_REFERENCE); call cke(sub,ret,79);
    !================================================================================
    do i=1,N_PROF; DATA_STATE_INDICATOR(i) = chr_DATA_STATE_INDICATOR; enddo
     ret=nf_put_var_text(ncid, DATA_STATE_IND_id, DATA_STATE_INDICATOR); call cke(sub,ret,80);
    !================================================================================
    do i=1,N_PROF; DATA_MODE(i) = chr_DATA_MODE; enddo ! ncprof - mode may be R or A
     ret=nf_put_var_text(ncid, DATA_MODE_id, DATA_MODE); call cke(sub,ret,81);
    !================================================================================
    do i=1,N_PROF; PLATFORM_TYPE(i) = chr_PLATFORM_TYPE; enddo
     ret=nf_put_var_text(ncid, PLATFORM_TYPE_id, PLATFORM_TYPE); call cke(sub,ret,45);
    !================================================================================
    if(profver3p0)then;
	do i=1,N_PROF; FLOAT_SERIAL_NO_3p0(i) = var_FLOAT_SERIAL_NO; enddo;
	 ret=nf_put_var_text(ncid, FLOAT_SERIAL_NO_id, FLOAT_SERIAL_NO_3p0); call cke(sub,ret,26);
    else; ! profver3p0 is F
	do i=1,N_PROF; FLOAT_SERIAL_NO(i) = var_FLOAT_SERIAL_NO; enddo
	 ret=nf_put_var_text(ncid, FLOAT_SERIAL_NO_id, FLOAT_SERIAL_NO); call cke(sub,ret,26);
    endif ! profver3p0
    !================================================================================
    if(profver3p0)then;
	do i=1,N_PROF; FIRMWARE_VERSION_3p0(i) = var_FIRMWARE_VERSION(1:16); enddo
	 ret=nf_put_var_text(ncid, FIRMWARE_VERSION_id, FIRMWARE_VERSION_3p0); call cke(sub,ret,53);
    else ! profver3p0
	do i=1,N_PROF; FIRMWARE_VERSION(i) = var_FIRMWARE_VERSION; enddo
	 ret=nf_put_var_text(ncid, FIRMWARE_VERSION_id, FIRMWARE_VERSION); call cke(sub,ret,53);
    endif ! profver3p0
    !================================================================================
    do i=1,N_PROF; WMO_INST_TYPE(i) = var_WMO_INST_TYPE; enddo
     ret=nf_put_var_text(ncid, WMO_INST_TYPE_id, WMO_INST_TYPE); call cke(sub,ret,82);
    !================================================================================
    do i=1,N_PROF; JULD(i) = prof_JULD; enddo
     ret=nf_put_var_double(ncid, JULD_id, JULD); call cke(sub,ret,83);
    !================================================================================
    do i=1,N_PROF; JULD_QC(i) = prof_JULD_QC; enddo
     ret=nf_put_var_text(ncid, JULD_QC_id, JULD_QC); call cke(sub,ret,84);
    !================================================================================
    do i=1,N_PROF; JULD_LOCATION(i) = prof_JULD_LOCATION; enddo
     ret=nf_put_var_double(ncid, JULD_LOCATION_id, JULD_LOCATION); call cke(sub,ret,85);
    !================================================================================
    do i=1,N_PROF; LATITUDE(i) = prof_LATITUDE; enddo
     ret=nf_put_var_double(ncid, LATITUDE_id, LATITUDE); call cke(sub,ret,86);
    !================================================================================
    do i=1,N_PROF; LONGITUDE(i) = prof_LONGITUDE; enddo
     ret=nf_put_var_double(ncid, LONGITUDE_id, LONGITUDE); call cke(sub,ret,87);
    !================================================================================
    do i=1,N_PROF; POSITION_QC(i) = prof_POSITION_QC; enddo
     ret=nf_put_var_text(ncid, POSITION_QC_id, POSITION_QC); call cke(sub,ret,88);
    !================================================================================
    do i=1,N_PROF; POSITIONING_SYSTEM(i) = chr_POSITIONING_SYSTEM; enddo
     ret=nf_put_var_text(ncid, POSITIONING_SYSTEM_id, POSITIONING_SYSTEM); call cke(sub,ret,89);
    !================================================================================
    do i=1,N_PROF; VERT_SAMPLING_SCHEME(i) = chr_VERT_SAMPLING_SCHEME(i); enddo;
     ret=nf_put_var_text(ncid, VERT_SAMPLING_SCHEME_id, VERT_SAMPLING_SCHEME); call cke(sub,ret,64);
    !================================================================================
    do i=1,N_PROF; CONFIG_MISSION_NUMBER(i) = int_CONFIG_MISSION_NUMBER; enddo
     ret=nf_put_var_int(ncid, CONFIG_MISSION_NBR_id, CONFIG_MISSION_NUMBER); call cke(sub,ret,76);
    !================================================================================

end subroutine put_ncprof_vars

  !====================================================================
subroutine def_ncprof_params(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;
    integer :: loc = 0;
    integer :: ipar;
    integer :: TMP_ids(5);	! PAR PAR_QC PAR_ADJ PAR_ADJ_QC PAR_ERR

    character(len=16), parameter :: sub = 'def_ncprof_parms'

    N_PARAM = nparam;

    if(DBLEV>0)then;write(msg,*)'In ',sub,' hasNCPAR: ',hasNCPAR; call wrtdbg(sub,msg); endif

    ! NOTE - define the nc variable PROFILE_CNDC_QC, if appropriate ...
    do ipar=1,MAXPAR
	if(hasNCPAR(ipar))then
	    if(DBLEV>0)then; write(msg,'("Define PROFILE_<>_QC for ",a)')ncparname(ipar); call wrtdbg(sub,msg); endif
	    call def_ncprof_profile_qc(caller,ncid,ncparname(ipar),PROF_QC_ids(ipar));
	    if(DBLEV>0)then; write(msg,*)'after def_ncprof_profile_qc, PROF_QC_ids: ',PROF_QC_ids; call wrtdbg(sub,msg); endif
	else
	    if(DBLEV>0)then;
		write(msg,'("Do NOT define PROFILE_<>_QC for ipar: ",i2,", var: ",a)')ipar,ncparname(ipar);
		call wrtdbg(sub,msg);
	    endif
	endif
    enddo

    do ipar=1,MAXPAR
	if(hasNCPAR(ipar))then
	    if(DBLEV>0)then;
		write(msg,'("Define PARAMETER array for ipar: ",i2," par: ",a)')ipar,ncparname(ipar);
		call wrtdbg(sub,msg);
	    endif
	    call def_ncprof_param(caller,ncid,ncparname(ipar),TMP_ids);
	    call def_ncprof_param_att(sub,ncid,ipar,ncparname(ipar),TMP_ids);
	    do inca=1,5
		VAR_ids(inca,ipar) = TMP_ids(inca)
	    enddo
	    write(msg,*)'TMPids @1 ',TMP_ids; call wrtdbg(sub,msg);
	else
	    if(DBLEV>0)then; write(msg,'("Do NOT define PARAMETER array for ipar: ",i2," par: ",a)')ipar,ncparname(ipar); endif
	endif
    enddo

    if(hasDOXY(3))then;write(msg,'("W DOXYA will NOT be written to profile nc file")');call wrtdbg(sub,msg);endif
    if(hasDOXY(4))then;write(msg,'("W DOXYB will NOT be written to profile nc file")');call wrtdbg(sub,msg);endif
    if(hasDOXY(5))then;write(msg,'("W DOXYC will NOT be written to profile nc file")');call wrtdbg(sub,msg);endif

    if(DBLEV>0)then; write(msg,'("<< Leave >>")'); call wrtdbg(sub,msg); endif

    !==================================================================

end subroutine def_ncprof_params

  !====================================================================
subroutine put_ncprof_params(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;
    integer :: loc = 0;
    integer :: ipar
    integer :: TMP_ids(5);      ! PAR PAR_QC PAR_ADJ PAR_ADJ_QC PAR_ERR

    character(len=16), parameter :: sub = 'put_ncprof_parms'

    if(DBLEV>0)then; write(msg,'("<< Enter ",a," >>")')sub; call wrtdbg(sub,msg); endif

    ! 20180308 - The final decision is to use the PSAL_QC for DNCD_QC, also PROFILE_PSAL_QC for PROFILE_CNDC_QC
    do ipar=1,MAXPAR
	if(hasNCPAR(ipar))then
	    if(ncparname(ipar) == "CNDC")then
		call put_ncprof_profile_qc(caller,ncid,ncparname(ipar),PROF_QC_ids(ipar));
	    else
		if(DBLEV>0)then; write(msg,'("Put PROFILE_<>_QC for ",a)')ncparname(ipar); call wrtdbg(sub,msg); endif
		call put_ncprof_profile_qc(caller,ncid,ncparname(ipar),PROF_QC_ids(ipar));
	    endif
	endif
    enddo

    if(adjustments)then
	if(.not. do_pres_adj)then
	    do ilev=1,nlev
		profdata(p1a,ilev) = profdata(p1r,ilev); profflag(p1a,ilev) = profflag(p1r,ilev);
		if(hasPRES(2))then; profdata(p2a,ilev) = profdata(p2r,ilev); profflag(p2a,ilev) = profflag(p2r,ilev); endif
		if(hasPRES(3))then; profdata(p3a,ilev) = profdata(p3r,ilev); profflag(p3a,ilev) = profflag(p3r,ilev); endif
		if(hasPRES(4))then; profdata(p4a,ilev) = profdata(p4r,ilev); profflag(p4a,ilev) = profflag(p4r,ilev); endif
	    enddo
	endif
	if(.not. do_psal_adj)then
	    do ilev=1,nlev
		profdata(s1a,ilev) = profdata(s1r,ilev); profflag(s1a,ilev) = profflag(s1r,ilev);
		if(hasPSAL(2))then; profdata(s2a,ilev) = profdata(s2r,ilev); profflag(s2a,ilev) = profflag(s2r,ilev); endif
		if(hasPSAL(3))then; profdata(s3a,ilev) = profdata(s3r,ilev); profflag(s3a,ilev) = profflag(s3r,ilev); endif
		if(hasPSAL(4))then; profdata(s4a,ilev) = profdata(s4r,ilev); profflag(s4a,ilev) = profflag(s4r,ilev); endif
            enddo
        endif
    endif

    do ipar=1,MAXPAR
	if(hasNCPAR(ipar))then
	    do inca=1,5; TMP_ids(inca) = VAR_ids(inca,ipar); enddo
	    if(DEBUG) write(DBG,'("PUTS calls PUT for ipar: ",i2," PAR: ",a," ids: ",5i3)')ipar,ncparname(ipar),TMP_ids
	    call put_ncprof_param(sub,ncid,ipar,ncparname(ipar),TMP_ids)
	else
	    if(DEBUG) write(DBG,'("PUTS skips PUT for ipar: ",i2," PAR: ",a," hasNCPAR=F")')ipar,ncparname(ipar)
	endif
	if(DEBUG) call wrtdbg(sub,msg);
    enddo
    if(DBLEV>0)then; write(msg,'("<< Leave ",a," >>")')sub; call wrtdbg(sub,msg); endif

    !==================================================================

end subroutine put_ncprof_params

  !====================================================================
subroutine def_ncprof_profile_qc(caller,ncid,PAR,PAR_id);
    character(*), intent( in )     :: caller
    integer, intent( in )          :: ncid 	! LNetCDF file id
    character(*), intent( in )      :: PAR	! PARAMETER to define
    integer, intent(out )          :: PAR_id;

    integer :: DATA_dim;
    character(len=48) :: nam
    integer :: lst;

    integer :: iret
    integer :: ret = 0;
    integer ::  loc
    character(len=18), parameter   :: sub = 'ncp_def_profile_qc';

    DATA_dim = N_PROF_dimid;

    nam="PROFILE_" // trim(PAR)//"_QC";  ret=nf_def_var(ncid,nam,NF_CHAR,NDIMS1,DATA_dim,PAR_id); call cke(sub,ret,1);

    nam="Global quality flag of " // trim(PAR) // " profile"; lst = len(trim(nam))
     ret=nf_put_att_text(ncid, PAR_id, 'long_name', lst, nam); call cke(sub,ret, 2);
     ret=nf_put_att_text(ncid, PAR_id, 'conventions', 23, 'Argo reference table 2a")'); call cke(sub,ret, 3);
     ret=nf_put_att_text(ncid, PAR_id, '_FillValue', 1, ' '); call cke(sub,ret, 4);

end subroutine def_ncprof_profile_qc

  !====================================================================
subroutine put_ncprof_profile_qc(caller,ncid,PAR,PAR_id);
    character(*), intent( in )     :: caller
    integer, intent( in )          :: ncid      ! LNetCDF file id
    character(*), intent( in )     :: PAR      ! PARAMETER to load (only used for annnotation)
    integer, intent( in )          :: PAR_id;

    integer :: iret
    integer :: ret = 0;
    integer :: loc
    integer          :: ipar
    character(len=1) :: grade(MAXPAR) = (/ "x", "x", "x", "x", "x" /);
    character(len=18), parameter   :: sub = 'put_ncp_profile_qc';

    if(DBLEV>0)then; write(msg,'("PR_QC: Put profile_qc for ",a)')PAR; call wrtdbg(sub,msg); endif

    ! Note that these refer to the DATA order (ie, profdata, etc.) notthe NCPAR order
        if(trim(PAR) == "PRES")then; ipar = 1;
    elseif(trim(PAR) == "TEMP")then; ipar = 2;
    elseif(trim(PAR) == "PSAL")then; ipar = 3;
    elseif(trim(PAR) == "CNDC")then; ipar = 3;
    elseif(trim(PAR) == "DOXY")then; ipar = 4;
    else
	write(msg,'("PR_QC: put_ncprof_profile_qc SHOULD NOT REACH THIS POINT")')
    endif

    grade(1) = " "; ! nc file missing vlue flag - use for nprf_stat >1
    do iprf=1,nprf
	if(PAR == "PRES")then;
	    i = iprf; 
	    grade(iprf) = prof_pres_qc(iprf)
	elseif(PAR == "TEMP")then;
	    grade(iprf) = prof_temp_qc(iprf)
	elseif(PAR == "PSAL")then;
	    grade(iprf) = prof_psal_qc(iprf)
	elseif(PAR == "CNDC")then;				            ! 20180308 use PSAL flag grade for CNDC
	    if(hasCNDC(iprf))then
		grade(iprf) = prof_psal_qc(iprf); ! Only if the profile HAS CNDC
	    else
		grade(iprf) = ' '; ! Otherwide, ncfillvalue
	     endif
	elseif(PAR == "DOXY")then;
	    grade(iprf) = prof_doxy_qc(iprf)
	endif
	if(grade(iprf) == 'M')then;
	    grade(iprf) = " "; ! nc file missing vlue flag
	elseif(grade(iprf) == 'M')then;
	    grade(iprf) = " "; ! nc file missing vlue flag
	elseif(grade(iprf) == 'Z')then;
	     grade(iprf) = " "; ! nc file missing vlue flag
	endif
    enddo

    if(DBLEV>0)then;
	write(msg,*)'NC_PRQC: put_ncprof_profile_qc PAR: ',PAR,' PAR_id: ',PAR_id,' grade: ',grade;
	call wrtdbg(sub,msg);
    endif
     ret=nf_put_var_text(ncid, PAR_id, grade);    call cke(sub,ret,1);
end subroutine put_ncprof_profile_qc

  !====================================================================pres PRES
subroutine def_ncprof_param(caller,ncid,PAR,PAR_ids);
character(*), intent( in )     :: caller
integer, intent( in )          :: ncid      ! LNetCDF file id
character(*), intent( in )      :: PAR      ! PARAMETER to define
integer, intent(out )          :: PAR_ids(5);

integer :: DATA_dims(NDIMS2);
character(len=48) :: nam

integer :: iret
integer :: ret = 0;
integer :: loc
character(len=16), parameter   :: sub = 'def_ncprof_param';

DATA_dims(2)         = N_PROF_dimid;  DATA_dims(1)         = N_LEVELS_dimid;

nam=trim(PAR);                     ret=nf_def_var(ncid,nam,NF_REAL,NDIMS2,DATA_dims,PAR_ids(1)); call cke(sub,ret,1);
nam=trim(PAR)//"_QC";              ret=nf_def_var(ncid,nam,NF_CHAR,NDIMS2,DATA_dims,PAR_ids(2)); call cke(sub,ret,2);
nam=trim(PAR)//"_ADJUSTED";        ret=nf_def_var(ncid,nam,NF_REAL,NDIMS2,DATA_dims,PAR_ids(3)); call cke(sub,ret,3);
nam=trim(PAR)//"_ADJUSTED_QC";     ret=nf_def_var(ncid,nam,NF_CHAR,NDIMS2,DATA_dims,PAR_ids(4)); call cke(sub,ret,4);
nam=trim(PAR)//"_ADJUSTED_ERROR";  ret=nf_def_var(ncid,nam,NF_REAL,NDIMS2,DATA_dims,PAR_ids(5)); call cke(sub,ret,5);

end subroutine def_ncprof_param

  !====================================================================
subroutine def_ncprof_param_att(caller,ncid,ipar,PAR,PAR_ids);
    character(*), intent( in )     :: caller
    integer, intent( in )          :: ncid      ! LNetCDF file id
    integer, intent( in )          :: ipar	! parameter number
    character(*), intent( in )      :: PAR      ! PARAMETER to define
    integer, intent(out )          :: PAR_ids(5);

    real :: vmin(MAXPAR), vmax(MAXPAR);
    character(len=48) :: nam
    character(len=14) :: unit; integer :: lunit;	! unit name, length of unit name
    character(len=86) :: lnam; integer :: llnam
    character(len=86) :: lnamv, lnama, lname; integer :: llnamv, llnama, llname
    character(len=42) :: snam; integer :: lsnam

    character(len=6) :: cfmtval; integer :: cfmtlen
    character(len=6) :: ffmtval; integer :: ffmtlen

    character(len=1)  :: PAR_qcfl = ' ';

    integer :: iret
    integer :: ret = 0;
    integer ::  loc
    character(len=17), parameter   :: sub = 'ncp_def_param_att';

    unit = PAR_unit(ipar);   lunit = len(trim(unit));
   
    do i=1,(MAXPAR)
	vmin(i) = PAR_vmin(i);
	if(profver3p0) vmin(i) = PAR_vmin_3p0(i);
	vmax(i) = PAR_vmax(i);
	if(profver3p0) vmax(i) = PAR_vmax_3p0(i);
    enddo
    lnamv = PAR_lnamv(ipar); llnamv = len(trim(lnamv));
    lnama = PAR_lnama(ipar); llnama = len(trim(lnama));
    lname = PAR_lname(ipar); llname = len(trim(lname));
    if(profver3p0)then                                     ! profver3p0
	lnamv = PAR_lnam(ipar); llnamv = len(trim(lnamv)); ! profver3p0
	lnama = PAR_lnam(ipar); llnama = len(trim(lnama)); ! profver3p0
	lname = PAR_lnam(ipar); llname = len(trim(lname)); ! profver3p0
    endif                                                  ! profver3p0
    snam = PAR_snam(ipar);   lsnam = len(trim(snam));

    ! First, select the correct C_format &  FORTRAN_format
    if(trans_type == 'ARGOS') then
	cfmtval = DEFCFORARGOS(ipar);   cfmtlen = len(trim(cfmtval));
	ffmtval = DEFFFORARGOS(ipar);   ffmtlen = len(trim(ffmtval));
    else
	cfmtval = DEFCFORIRIDIUM(ipar); cfmtlen = len(trim(cfmtval));
	ffmtval = DEFFFORIRIDIUM(ipar); ffmtlen = len(trim(ffmtval));
    endif

    ! attributes for PAR

     ret=nf_put_att_text(ncid, PAR_ids(1), 'long_name',     llnamv, trim(lnamv));
     ret=nf_put_att_text(ncid, PAR_ids(1), 'standard_name', lsnam, trim(snam));
     ret=nf_put_att_real(ncid, PAR_ids(1), '_FillValue',    NF_REAL, 1, PAR_fill(ipar));
     ret=nf_put_att_text(ncid, PAR_ids(1), 'units',         lunit, trim(unit));
     ret=nf_put_att_real(ncid, PAR_ids(1), 'valid_min',     NF_REAL, 1,vmin(ipar));
     ret=nf_put_att_real(ncid, PAR_ids(1), 'valid_max',     NF_REAL, 1,vmax(ipar));
     ret=nf_put_att_text(ncid, PAR_ids(1), 'C_format',      cfmtlen, cfmtval);
     ret=nf_put_att_text(ncid, PAR_ids(1), 'FORTRAN_format',ffmtlen, ffmtval);
     ret=nf_put_att_real(ncid, PAR_ids(1), 'resolution',    NF_REAL, 1, DEFRSLN(ipar));
    if(PAR=="PRES")  ret=nf_put_att_text(ncid, PAR_ids(1), 'axis',1, 'Z')


    ! attributes for PAR_qc

     ret=nf_put_att_text(ncid, PAR_ids(2), 'long_name',   12, 'quality flag' );
     ret=nf_put_att_text(ncid, PAR_ids(2), 'conventions', 22, 'Argo reference table 2' );
     ret=nf_put_att_text(ncid, PAR_ids(2), '_FillValue',   1, ' ');

    ! attributes for PAR_adjusted

     ret=nf_put_att_text(ncid, PAR_ids(3), 'long_name',     llnama, trim(lnama));
     ret=nf_put_att_text(ncid, PAR_ids(3), 'standard_name', lsnam, trim(snam));
     ret=nf_put_att_real(ncid, PAR_ids(3), '_FillValue',    NF_REAL, 1, PAR_fill(ipar));
     ret=nf_put_att_text(ncid, PAR_ids(3), 'units',         lunit, trim(unit));
     ret=nf_put_att_real(ncid, PAR_ids(3), 'valid_min',     NF_REAL, 1,vmin(ipar));
     ret=nf_put_att_real(ncid, PAR_ids(3), 'valid_max',     NF_REAL, 1,vmax(ipar));
     ret=nf_put_att_text(ncid, PAR_ids(3), 'C_format',      cfmtlen, cfmtval);
     ret=nf_put_att_text(ncid, PAR_ids(3), 'FORTRAN_format',ffmtlen, ffmtval);
     ret=nf_put_att_real(ncid, PAR_ids(3), 'resolution',    NF_REAL, 1, DEFRSLN(ipar));

    ! attributes for PAR_adjusted_qc

     ret=nf_put_att_text(ncid, PAR_ids(4), 'long_name',   12, 'quality flag' );
     ret=nf_put_att_text(ncid, PAR_ids(4), 'conventions', 22, 'Argo reference table 2' );
     ret=nf_put_att_text(ncid, PAR_ids(4), '_FillValue',   1, ' ');


    ! attributes for PAR_error

     ret=nf_put_att_text(ncid, PAR_ids(5), 'long_name',     llname, trim(lname));
     ret=nf_put_att_real(ncid, PAR_ids(5), '_FillValue',    NF_REAL, 1, PAR_fill(ipar));
     ret=nf_put_att_text(ncid, PAR_ids(5), 'units',         lunit, trim(unit));
     ret=nf_put_att_text(ncid, PAR_ids(5), 'C_format',      cfmtlen, cfmtval);
     ret=nf_put_att_text(ncid, PAR_ids(5), 'FORTRAN_format',ffmtlen, ffmtval);
     ret=nf_put_att_real(ncid, PAR_ids(5), 'resolution',    NF_REAL, 1, DEFRSLN(ipar));

end subroutine def_ncprof_param_att

  !====================================================================
subroutine put_ncprof_param(caller,ncid,ipar,PAR,PAR_ids);


    character(*), intent( in )     :: caller
    integer, intent( in )          :: ncid       ! LNetCDF file id
    integer, intent( in )          :: ipar       ! PARAMETER number
    character(*), intent( in )      :: PAR       ! PARAMETER name to load (only used for annnotation)
    integer, intent( in )          :: PAR_ids(5) ! PARAMETER nc ids;

    integer :: iret
    integer :: ret = 0;
    integer :: loc
    integer :: ireg(MAXPRF), iadj(MAXPRF);
    integer :: count(2), start(2)
    real    :: DATAREG(MAXLEV); character(len=1) :: FLAGREG(MAXLEV);
    real    :: DATAADJ(MAXLEV); character(len=1) :: FLAGADJ(MAXLEV);

    character(len=16), parameter   :: sub = 'put_ncprof_param';

    if(DBLEV /= 37)then;
	if(DEBUG)then; write(msg,'(a," load param number ",i2," variable: ",a)')sub,ipar,PAR; call wrtdbg(sub,msg); endif
    endif

    count(1) = nlev; count(2) = 1;
    start(1) = 1;    start(2) = 1;	! Note that start(2) is the profile number (iprf)

    if(PAR == "CNDC")then
	if(DEBUG)then; write(msg,*)'CNDC ALLPRF: hasCNDC: ',hasCNDC; call wrtdbg(sub,msg); endif
	do iprf = 1,N_PROF
	    ! Initialize/fluxh the nc arrays
	    do ilev=1,nlev; DATAREG(ilev)=FILLNC; DATAADJ(ilev)=FILLNC; FLAGREG(ilev)=' '; FLAGADJ(ilev)=' '; enddo
	    if(hasCNDC(iprf))then
		do ilev=1,nlev
		    ! 01Mar2018 - use the fill value (space), not the :was not tested (zero) for CNDC
		    if(profcndc(iprf,ilev)/=FILL)then
                        ! 20180308 - The final decision is to use the PSAL_QC for DNCD_QC
			if(profflag(regPSAL(iprf),ilev) < 5)then
			    ! 20180524 Do NOT output interpolated data into the nc file
                            DATAREG(ilev)=profcndc(iprf,ilev); FLAGREG(ilev)=ncflag(profflag(regPSAL(iprf),ilev));
                            if(cnvrt_cndc(iprf))then; DATAREG(ilev)=DATAREG(ilev)* multplr_CM2M; endif;
			    if(adjustments)then
				DATAADJ(ilev)=profcndc(iprf,ilev); FLAGADJ(ilev)=ncflag(profflag(adjPSAL(iprf),ilev));
                                if(cnvrt_cndc(iprf))then; DATAADJ(ilev)=DATAADJ(ilev)* multplr_CM2M; endif;
			    endif
			endif
		    endif
		enddo
	    endif

	    start(2) = iprf;
	     ret=nf_put_vara_real(ncid, PAR_ids(1), start, count, DATAREG);  call cke(sub,ret,201);
	    if(DBLEV==37)then; write(msg,1001)PAR,'DATAREG',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	     ret=nf_put_vara_text(ncid, PAR_ids(2), start, count, FLAGREG);  call cke(sub,ret,202);
	    if(DBLEV==37)then; write(msg,1001)PAR,'FLAGREG',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	    if(adjustments)then
		 ret=nf_put_vara_real(ncid, PAR_ids(3), start, count, DATAADJ);  call cke(sub,ret,203);
		if(DBLEV==37)then; write(msg,1001)PAR,'DATAADJ',iprf,ret,start,count; call wrtdbg(sub,msg); endif
		 ret=nf_put_vara_text(ncid, PAR_ids(4), start, count, FLAGADJ);  call cke(sub,ret,204);
		if(DBLEV==37)then; write(msg,1001)PAR,'FLAGADJ',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	    else
		 ret=nf_put_vara_real(ncid, PAR_ids(3), start, count, FILLNCD);  call cke(sub,ret,203);
		if(DBLEV==37)then; write(msg,1001)PAR,'DATAADJ',iprf,ret,start,count; call wrtdbg(sub,msg); endif
		 ret=nf_put_vara_text(ncid, PAR_ids(4), start, count, FILLNCF);  call cke(sub,ret,204);
		if(DBLEV==37)then; write(msg,1001)PAR,'FLAGADJ',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	    endif
	     ret=nf_put_vara_real(ncid, PAR_ids(5), start, count, FILLNCD);  call cke(sub,ret,205);
	    if(DBLEV==37)then; write(msg,1001)PAR,'DATAERR',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	enddo
	if(DBLEV==37)then; write(msg,'("The CNDC arrays have been loaded, reurrn to the caller")'); call wrtdbg(sub,msg); endif
	return
    endif

    ! I know, I know, total kluge
    ireg(1) = 0; ireg(2) = 0;  ireg(3) = 0; ireg(4) = 0;
    iadj(1) = 0; iadj(2) = 0;  iadj(3) = 0; iadj(4) = 0;
    if(PAR == "PRES")then
	if(hasPRES(1))then; ireg(1) = p1r; if(adjustments)then; iadj(1) = p1a; endif; endif
	if(hasPRES(2))then; ireg(2) = p2r; if(adjustments)then; iadj(2) = p2a; endif; endif
	if(hasPRES(3))then; ireg(3) = p3r; if(adjustments)then; iadj(3) = p3a; endif; endif
	if(hasPRES(4))then; ireg(4) = p4r; if(adjustments)then; iadj(4) = p4a; endif; endif
    elseif(PAR == "TEMP")then
	if(hasTEMP(1))then; ireg(1) = t1r; if(adjustments)then; iadj(1) = t1r; endif; endif
	if(hasTEMP(2))then; ireg(2) = t2r; if(adjustments)then; iadj(2) = t2r; endif; endif
	if(hasTEMP(3))then; ireg(3) = t3r; if(adjustments)then; iadj(3) = t3r; endif; endif
	if(hasTEMP(4))then; ireg(4) = t4r; if(adjustments)then; iadj(4) = t4r; endif; endif
    elseif(PAR == "PSAL")then
	if(hasPSAL(1))then; ireg(1) = s1r; if(adjustments)then; iadj(1) = s1a; endif; endif
	if(hasPSAL(2))then; ireg(2) = s2r; if(adjustments)then; iadj(2) = s2a; endif; endif
	if(hasPSAL(3))then; ireg(3) = s3r; if(adjustments)then; iadj(3) = s3a; endif; endif
	if(hasPSAL(4))then; ireg(4) = s4r; if(adjustments)then; iadj(4) = s4a; endif; endif
    elseif(PAR == "DOXY")then
	if(outRDOXY .and. hasDOXY(1))then; ireg(1) = do1; if(adjustments)then; iadj(1) = do1; endif; endif
	if(outRDOXY .and. hasDOXY(2))then; ireg(2) = do2; if(adjustments)then; iadj(2) = do2; endif; endif ! doxy2  is nprf=2
	if(outRDOXY .and. hasDOXY(3))then; ireg(2) = doa; if(adjustments)then; iadj(2) = doa; endif; endif ! doxy2a is nprf=2
    endif
    ! for thos arrays that are fill/missing value, use FILLNCD & FILLNCF
    do iprf = 1,N_PROF
	! Initialize/fluxh the nc arrays
	do ilev=1,nlev; DATAREG(ilev)=FILLNC; DATAADJ(ilev)=FILLNC; FLAGREG(ilev)=' '; FLAGADJ(ilev)=' '; enddo
	do ilev = 1,nlev
	    DATAREG(ilev)=FILLNC; DATAADJ(ilev)=FILLNC; FLAGREG(ilev)=' '; FLAGADJ(ilev)=' ';
	    if(ireg(iprf) > 0)then
		if(profdata(ireg(iprf),ilev)/=FILL)then
		    if(profflag(ireg(iprf),ilev) < 5)then
			! 20180524 Do NOT output interpolated data into the nc file
			DATAREG(ilev)=profdata(ireg(iprf),ilev);
			FLAGREG(ilev) = ncflag(profflag(ireg(iprf),ilev));
			if(adjustments)then
			    DATAADJ(ilev)=profdata(iadj(iprf),ilev);
			    FLAGADJ(ilev) = ncflag(profflag(iadj(iprf),ilev));
			endif
		    endif
		endif
	    endif
	enddo

 	10 continue ! Skip the standard way comes here

	start(2) = iprf;
	 ret=nf_put_vara_real(ncid, PAR_ids(1), start, count, DATAREG);  call cke(sub,ret,201);
	if(DBLEV==37)then; write(msg,1001)PAR,'DATAREG',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	 ret=nf_put_vara_text(ncid, PAR_ids(2), start, count, FLAGREG);  call cke(sub,ret,202);
	if(DBLEV==37)then; write(msg,1001)PAR,'FLAGREG',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	if(adjustments)then
	     ret=nf_put_vara_real(ncid, PAR_ids(3), start, count, DATAADJ);  call cke(sub,ret,203);
	    if(DBLEV==37)then; write(msg,1001)PAR,'DATAADJ',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	     ret=nf_put_vara_text(ncid, PAR_ids(4), start, count, FLAGADJ);  call cke(sub,ret,204);
	    if(DBLEV==37)then; write(msg,1001)PAR,'FLAGADJ',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	else
	     ret=nf_put_vara_real(ncid, PAR_ids(3), start, count, FILLNCD);  call cke(sub,ret,203);
	    if(DBLEV==37)then; write(msg,1001)PAR,'DATAADJ',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	     ret=nf_put_vara_text(ncid, PAR_ids(4), start, count, FILLNCF);  call cke(sub,ret,204);
	    if(DBLEV==37)then; write(msg,1001)PAR,'FLAGADJ',iprf,ret,start,count; call wrtdbg(sub,msg); endif
	endif
	 ret=nf_put_vara_real(ncid, PAR_ids(5), start, count, FILLNCD);  call cke(sub,ret,205);
	if(DBLEV==37)then; write(msg,1001)PAR,'DATAERR',iprf,ret,start,count; call wrtdbg(sub,msg); endif

    enddo
    1001 format("NEW put_param PUT ",a," ",a," iprf: ",i2," iret: ",i4," start/count: ",4i4)

end subroutine put_ncprof_param

  !====================================================================
  ! SCIENTIFIC_CALIB defs
  !====================================================================
subroutine def_ncprof_calhis(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;

    character(len=64) :: long_name;
    character(len=70) :: conventns;
    integer :: nn, nc, nu;  ! length of long_name  & conventions

    character(len=16), parameter :: sub = 'def_ncprof_calhis'

    if(DBLEV==37)then; write(msg,'("LEV37 Enter sub ",a)')sub; call wrtdbg(sub,msg); endif

    !===================================================================
    SCICAL_dims(4) = N_PROF_dimid;
    SCICAL_dims(3) = N_CALIB_dimid;
    SCICAL_dims(2) = N_PARAM_dimid;
    !-------------------------------------------------------------------
    SCICAL_dims(1) = STRING16_dimid; ! ncprof PARAMETER names are 16 for coreArgo
    !-------------------------------------------------------------------
    if(DBLEV==37)then; write(msg,*)'LEV37 BEFORE def, SCICAL_ids: ',SCICAL_ids; call wrtdbg(sub,msg); endif
     ret=nf_def_var(ncid, 'PARAMETER',                    NF_CHAR,NDIMS4, SCICAL_dims, SCICAL_ids(1)); call cke(sub,ret,11);
    if(DBLEV==37)then; write(msg,*)'LEV37 AFTER  def, SCICAL_ids: ',SCICAL_ids; call wrtdbg(sub,msg); endif
    long_name = "List of parameters with calibration information";                       nn = len(trim(long_name))
    conventns = "Argo reference table 3";                                                nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, SCICAL_ids(1), 'long_name',   nn, long_name);           call cke(sub,ret,12);
     ret=nf_put_att_text(ncid, SCICAL_ids(1), 'conventions', nc, conventns);           call cke(sub,ret,13);
     ret=nf_put_att_text(ncid, SCICAL_ids(1), '_FillValue', 1, ' ');                   call cke(sub,ret,14);
    !-------------------------------------------------------------------
    SCICAL_dims(1) = STRING256_dimid;
    !-------------------------------------------------------------------
     ret=nf_def_var(ncid, 'SCIENTIFIC_CALIB_EQUATION',    NF_CHAR,NDIMS4, SCICAL_dims, SCICAL_ids(2)); call cke(sub,ret,21);
    long_name = "Calibration equation for this parameter";                               nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, SCICAL_ids(2), 'long_name',   nn, long_name);           call cke(sub,ret,22);
     ret=nf_put_att_text(ncid, SCICAL_ids(2), '_FillValue', 1, ' ');                   call cke(sub,ret,23);
    !-------------------------------------------------------------------
     ret=nf_def_var(ncid, 'SCIENTIFIC_CALIB_COEFFICIENT', NF_CHAR,NDIMS4, SCICAL_dims, SCICAL_ids(3)); call cke(sub,ret,31);
    long_name = "Calibration coefficients for this equation";                             nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, SCICAL_ids(3), 'long_name',   nn, long_name);           call cke(sub,ret,32);
     ret=nf_put_att_text(ncid, SCICAL_ids(3), '_FillValue', 1, ' ');                   call cke(sub,ret,33);
    !-------------------------------------------------------------------
     ret=nf_def_var(ncid, 'SCIENTIFIC_CALIB_COMMENT',     NF_CHAR,NDIMS4, SCICAL_dims, SCICAL_ids(4)); call cke(sub,ret,41);
    long_name = "Comment applying to this parameter calibration";                        nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, SCICAL_ids(4), 'long_name',   nn, long_name);           call cke(sub,ret,42);
     ret=nf_put_att_text(ncid, SCICAL_ids(4), '_FillValue', 1, ' ');                   call cke(sub,ret,43);
    !-------------------------------------------------------------------
    SCICAL_dims(1) = DATE_TIME_dimid;
    !-------------------------------------------------------------------
     ret=nf_def_var(ncid, 'SCIENTIFIC_CALIB_DATE',        NF_CHAR,NDIMS4, SCICAL_dims,SCICAL_ids(5)); call cke(sub,ret,51);
    long_name = "Date of calibration"; nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, SCICAL_ids(5), 'long_name',   nn, long_name);           call cke(sub,ret,52);
     ret=nf_put_att_text(ncid, SCICAL_ids(5), 'conventions',  14,'YYYYMMDDHHMISS');
     ret=nf_put_att_text(ncid, SCICAL_ids(5), '_FillValue', 1, ' ');                   call cke(sub,ret,53);

    !================================================================================
    ! HISTORY defs
    !================================================================================
    ! We need both HIST_dim2 and HIST_DIM3 - there are 12 history variables - three of them
    ! are REALs, and have TWO dimentions:
    !    float HISTORY_START_PRES      (N_HISTORY, N_PROF) ;
    !    float HISTORY_STOP_PRES       (N_HISTORY, N_PROF) ;
    !    float HISTORY_PREVIOUS_VALUE  (N_HISTORY, N_PROF) ;
    ! and the rest are character, and have a third dimensiont (size of the char string):
    !    char HISTORY_INSTITUTION      (N_HISTORY, N_PROF, STRING4) ;
    !    char HISTORY_STEP             (N_HISTORY, N_PROF, STRING4) ;
    !    char HISTORY_SOFTWARE         (N_HISTORY, N_PROF, STRING4) ;
    !    char HISTORY_SOFTWARE_RELEASE (N_HISTORY, N_PROF, STRING4) ;
    !    char HISTORY_REFERENCE        (N_HISTORY, N_PROF, STRING64) ;
    !    char HISTORY_DATE             (N_HISTORY, N_PROF, DATE_TIME) ;
    !    char HISTORY_ACTION           (N_HISTORY, N_PROF, STRING4) ;
    !    char HISTORY_PARAMETER        (N_HISTORY, N_PROF, STRING16) ; ! ncprof PARAMETER names are 16 for coreArgo
    !    char HISTORY_QCTEST           (N_HISTORY, N_PROF, STRING16) ;

    HIST_dim3(3) = N_HISTORY_dimid ! for all the 3D variables (STRING*nn is the first element)
    HIST_dim3(2) = N_PROF_dimid    ! for all the 3D variables
    HIST_dim2(2) = N_HISTORY_dimid ! for all the 2D variables
    HIST_dim2(1) = N_PROF_dimid    ! for all the 2D variables
    ! 01c-----------------------------------------------------------------
    HIST_dim3(1) = STRING4_dimid;  ! char HISTORY_INSTITUTION      (N_HISTORY, STRING4);
     ret=nf_def_var(ncid, 'HISTORY_INSTITUTION',     NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 1));    call cke(sub,ret,1010);
    long_name = "Institution which performed action";                                               nn = len(trim(long_name))
    conventns = "Argo reference table 4";                                                           nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 1), 'long_name',   nn, long_name);                       call cke(sub,ret,1011);
     ret=nf_put_att_text(ncid, HIST_ids( 1), 'conventions', nc, conventns);                       call cke(sub,ret,1012);
     ret=nf_put_att_text(ncid, HIST_ids( 1), '_FillValue',   1, ' ');                             call cke(sub,ret,1013);
    ! 02c-----------------------------------------------------------------
    HIST_dim3(1) = STRING4_dimid;  ! char HISTORY_STEP             (N_HISTORY, STRING4);
     ret=nf_def_var(ncid, 'HISTORY_STEP',            NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 2));    call cke(sub,ret,1020);
    long_name = "Step in data processing";                                                          nn = len(trim(long_name))
    conventns = "Argo reference table 12";                                                          nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 2), 'long_name',   nn, long_name);                       call cke(sub,ret,1021);
     ret=nf_put_att_text(ncid, HIST_ids( 2), 'conventions', nc, conventns);                       call cke(sub,ret,1022);
     ret=nf_put_att_text(ncid, HIST_ids( 2), '_FillValue',   1, ' ');                             call cke(sub,ret,1023);
    ! 03c-----------------------------------------------------------------
    HIST_dim3(1) = STRING4_dimid;  ! char HISTORY_SOFTWARE         (N_HISTORY, STRING4);
     ret=nf_def_var(ncid, 'HISTORY_SOFTWARE',        NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 3));    call cke(sub,ret,1030);
    long_name = "Name of software which performed action";                                          nn = len(trim(long_name))
    conventns = "Institution dependent";                                                            nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 3), 'long_name',   nn, long_name);                       call cke(sub,ret,1031);
     ret=nf_put_att_text(ncid, HIST_ids( 3), 'conventions', nc, conventns);                       call cke(sub,ret,1032);
     ret=nf_put_att_text(ncid, HIST_ids( 3), '_FillValue',   1, ' ');                             call cke(sub,ret,1033);
    ! 04c-----------------------------------------------------------------
    HIST_dim3(1) = STRING4_dimid;  ! char HISTORY_SOFTWARE_RELEASE (N_HISTORY, STRING4);
     ret=nf_def_var(ncid, 'HISTORY_SOFTWARE_RELEASE',NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 4));    call cke(sub,ret,1040);
    long_name = "Version/release of software which performed action";                               nn = len(trim(long_name))
    conventns = "Institution dependent";                                                            nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 4), 'long_name',   nn, long_name);                       call cke(sub,ret,1041);
     ret=nf_put_att_text(ncid, HIST_ids( 4), 'conventions', nc, conventns);                       call cke(sub,ret,1042);
     ret=nf_put_att_text(ncid, HIST_ids( 4), '_FillValue',   1, ' ');                             call cke(sub,ret,1043);
    ! 05c-----------------------------------------------------------------
    HIST_dim3(1) = STRING64_dimid; ! char HISTORY_REFERENCE        (N_HISTORY, STRING64);
     ret=nf_def_var(ncid, 'HISTORY_REFERENCE',       NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 5));    call cke(sub,ret,1050);
    long_name = "Reference of database";                                                            nn = len(trim(long_name))
    conventns = "Institution dependent";                                                            nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 5), 'long_name',   nn, long_name);                       call cke(sub,ret,1051);
     ret=nf_put_att_text(ncid, HIST_ids( 5), 'conventions', nc, conventns);                       call cke(sub,ret,1052);
     ret=nf_put_att_text(ncid, HIST_ids( 5), '_FillValue',   1, ' ');                             call cke(sub,ret,1053);
    ! 06d-----------------------------------------------------------------
    HIST_dim3(1) = DATE_TIME_dimid ! char HISTORY_DATE             (N_HISTORY, DATE_TIME);
     ret=nf_def_var(ncid, 'HISTORY_DATE',            NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 6));    call cke(sub,ret,1060);
    long_name = "Date the history record was created";                                              nn = len(trim(long_name))
    conventns = "YYYYMMDDHHMISS";                                                                   nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 6), 'long_name',   nn, long_name);                       call cke(sub,ret,1061);
     ret=nf_put_att_text(ncid, HIST_ids( 6), 'conventions', nc, conventns);                       call cke(sub,ret,1062);
     ret=nf_put_att_text(ncid, HIST_ids( 6), '_FillValue',   1, ' ');                             call cke(sub,ret,1063);
    ! 07c-----------------------------------------------------------------
    HIST_dim3(1) = STRING4_dimid;  ! char HISTORY_ACTION           (N_HISTORY, STRING4);
     ret=nf_def_var(ncid, 'HISTORY_ACTION',          NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 7));    call cke(sub,ret,1070);
    long_name = "Action performed on data";                                                         nn = len(trim(long_name))
    conventns = "Argo reference table 7";                                                           nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 7), 'long_name',   nn, long_name);                       call cke(sub,ret,1071);
     ret=nf_put_att_text(ncid, HIST_ids( 7), 'conventions', nc, conventns);                       call cke(sub,ret,1072);
     ret=nf_put_att_text(ncid, HIST_ids( 7), '_FillValue',   1, ' ');                             call cke(sub,ret,1073);
    ! 08c-----------------------------------------------------------------
    HIST_dim3(1) = STRING16_dimid; ! char HISTORY_PARAMETER        (N_HISTORY, STRING16); ! ncprof PARAMETER names are 16 for coreArgo
     ret=nf_def_var(ncid, 'HISTORY_PARAMETER',       NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 8));    call cke(sub,ret,1080);
    long_name = "Station parameter action is performed on";                                         nn = len(trim(long_name))
    conventns = "Argo reference table 3";                                                           nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 8), 'long_name',   nn, long_name);                       call cke(sub,ret,1081);
     ret=nf_put_att_text(ncid, HIST_ids( 8), 'conventions', nc, conventns);                       call cke(sub,ret,1082);
     ret=nf_put_att_text(ncid, HIST_ids( 8), '_FillValue',   1, ' ');                             call cke(sub,ret,1083);
    ! 09c-----------------------------------------------------------------
     ret=nf_def_var(ncid, 'HISTORY_START_PRES',     NF_REAL, NDIMS2, HIST_dim2, HIST_ids( 9));    call cke(sub,ret,1110);
    long_name = "Start pressure action applied on";                                                    nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, HIST_ids( 9), 'long_name',   nn, long_name);                       call cke(sub,ret,1111);
     ret=nf_put_att_real(ncid, HIST_ids( 9), '_FillValue',NF_REAL, 1, FILLNC);                     call cke(sub,ret,1113);
     ret=nf_put_att_text(ncid, HIST_ids( 9), 'units', 7, 'decibar');                               call cke(sub,ret,1123);
    ! 10i-----------------------------------------------------------------
     ret=nf_def_var(ncid, 'HISTORY_STOP_PRES',      NF_REAL, NDIMS2, HIST_dim2, HIST_ids(10));    call cke(sub,ret,1120);
    long_name = "Stop pressure action applied on";                                                     nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, HIST_ids(10), 'long_name',   nn, long_name);                       call cke(sub,ret,1121);
     ret=nf_put_att_real(ncid, HIST_ids(10), '_FillValue',NF_REAL, 1, FILLNC);                   call cke(sub,ret,1123);
     ret=nf_put_att_text(ncid, HIST_ids(10), 'units', 7, 'decibar');                               call cke(sub,ret,1123);
    ! 11r-----------------------------------------------------------------
     ret=nf_def_var(ncid, 'HISTORY_PREVIOUS_VALUE',  NF_REAL,NDIMS2, HIST_dim2, HIST_ids(11));    call cke(sub,ret,1090);
    long_name = "Parameter/Flag previous value before action";                                      nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, HIST_ids(11), 'long_name',   nn, long_name);                       call cke(sub,ret,1091);
     ret=nf_put_att_real(ncid, HIST_ids(11), '_FillValue',   NF_REAL, 1, FILLNC);                        call cke(sub,ret,1093);
    ! 12c-----------------------------------------------------------------
    HIST_dim3(1) = STRING16_dimid; ! char HISTORY_QCTEST           (N_HISTORY, STRING16);
     ret=nf_def_var(ncid, 'HISTORY_QCTEST',          NF_CHAR,NDIMS3, HIST_dim3, HIST_ids(12));    call cke(sub,ret,1130);
    long_name = "Documentation of tests performed, tests failed (in hex form)";                     nn = len(trim(long_name))
    conventns = "Write tests performed when ACTION=QCP$; tests failed when ACTION=QCF$";            nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids(12), 'long_name',   nn, long_name);                       call cke(sub,ret,1131);
     ret=nf_put_att_text(ncid, HIST_ids(12), 'conventions', nc, conventns);                       call cke(sub,ret,1132);
     ret=nf_put_att_text(ncid, HIST_ids(12), '_FillValue',   1, ' ');                             call cke(sub,ret,1133);
    !--------------------------------------------------------------------

    if(DBLEV==37)then; write(msg,'("LEV37 Leave sub ",a)')sub; call wrtdbg(sub,msg); endif

end subroutine def_ncprof_calhis
  !====================================================================
  ! SCIENTIFIC_CALIB puts
  !====================================================================
subroutine put_ncprof_calhis(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;
    integer ::  loc
    integer :: ihst, ipar, ical ! Counters for the dimensions
    character(len=14) :: c14;
    character(len=16) :: c16;
    character(len=64) :: c64;
    character(len=256) :: c256;		! SCICAL_<256 blank characters>
    character(len=256) :: b256;
    character(len=256) :: none;		! SCICAL_eqnt for unadjusted parameters


    character(len=16) :: perf_tmp, fail_tmp

    integer :: st2(2), ct2(2), st3(3), ct3(3), st4(4), ct4(4); ! start/count for vara uts
    ! integer :: st4(4), ct4(4); ! start/count for vara puts
    integer :: nhist
    integer :: ipf

    character(len=16) :: qcperf(MAXPRF), qcfail(MAXPRF);

    character(len=18), parameter :: sub = 'put_ncprof_calhis'

    if(DBLEV==37)then; write(msg,'("LEV37 Enter sub ",a)')sub; call wrtdbg(sub,msg); endif

    c14 = "              ";
    c16 = "                ";
    write(c64,'(a64)')' ';
    write(c256,'(a256)')' ';
    write(b256,'(a256)')' ';
    write(none,'("none",a252)')' ';

    !==================================================================
    N_PROF = nprf;
    !==================================================================
    do iprf = 1,N_PROF
	do ical = 1,N_CALIB
	    do ipar = 1,N_PARAM
		! BAD write(c16,'(a4,12" ")')ncparout(ipar);	!
		write(c16,'(a4,a12)')ncparout(ipar),' ';
		st4(1) = 1; !varies by string length
		st4(2) = ipar; ct4(2) = 1;	
		st4(3) =    1; ct4(3) = 1;	
		st4(4) = iprf; ct4(4) = 1;	
		write(PARAMETER,'(a4,a12)')ncparout(ipar),' ';
		if(DBLEV==37)then; write(msg,'("PUT_SCICAL PAR PARAMETER: >",a,"<")')PARAMETER; call wrtdbg(sub,msg); endif
		ct4(1) = 16;  ret=nf_put_vara_text(ncid, SCICAL_ids(1),st4,ct4,PARAMETER);         call cke(sub,ret,1);
		SCICAL_eqtn = none; SCICAL_coef = b256; SCICAL_cmnt = b256; SCICAL_date = c14;
		if(ncparout(ipar)=="PRES" .and. do_pres_adj)then
		    write(c256,'("PRES_ADJUSTED = PRES - surface_pressure")');       SCICAL_eqtn = adjustl(c256);
		    write(c256,'("surface_pressure=",a," dbar")')trim(cpres_offset); SCICAL_coef = adjustl(c256);
		    write(c256,'("Pressure adjusted at real time based on most recent valid surface pressure")');
		    SCICAL_cmnt = adjustl(c256);
		    SCICAL_date = file_date_time;
		elseif(ncparout(ipar)=="PSAL" .and. do_psal_adj)then
		    if(STS_prof(iprf))then;
			write(c256,'("PSAL_ADJUSTED not caluclated for STS profiles")'); ! STS_prof = .true.
			SCICAL_eqtn = adjustl(c256); SCICAL_cmnt = adjustl(c256);
		    else
			write(c256,'("PSAL_ADJUSTED = salinity + salinity_offset")'); SCICAL_eqtn = adjustl(c256);
			write(c256,'("salinity_offset = ",f10.7)')psal_offset; SCICAL_coef = adjustl(c256);
			write(c256,'("PSAL ADJUST [dd mm yyyy N S_off stddev] ",a)')trim(psal_adj_used);
			SCICAL_cmnt = adjustl(c256);
			c256 = trim(c256) // " where N is the number of the delayed-mode profile used to estimate S_off stddev";
			SCICAL_cmnt = adjustl(c256);
			SCICAL_date = file_date_time;
		    endif
		endif
		ct4(1) = 256;  ret=nf_put_vara_text(ncid, SCICAL_ids(2),st4,ct4,SCICAL_eqtn); call cke(sub,ret,2);
		ct4(1) = 256;  ret=nf_put_vara_text(ncid, SCICAL_ids(3),st4,ct4,SCICAL_coef); call cke(sub,ret,3);
		ct4(1) = 256;  ret=nf_put_vara_text(ncid, SCICAL_ids(4),st4,ct4,SCICAL_cmnt); call cke(sub,ret,4);
		ct4(1) = 14;   ret=nf_put_vara_text(ncid, SCICAL_ids(5),st4,ct4,SCICAL_date); call cke(sub,ret,5);
	    enddo
	enddo
    enddo
    !==================================================================
    ! HISTORY puts
    !================================================================================
    write(msg,'("!==================================================================")'); call wrtdbg(sub,msg);
    c16 = "                "; ! Clear after SCICAL puts
    ! These are the same for all history records
    NHIST_inst = "AO  ";		! HISTORY_INSTITUTION
    NHIST_sfrl = "    ";		! HISTORY_SOFTWARE_RELEASE
    NHIST_refr = c64;			! HISTORY_REFERENCE
    NHIST_date = file_date_time;	! HISTORY_DATE
    NHIST_parm = "                ";	! HISTORY_PARAMETER
    NHIST_strt = FILLNC;		! HISTORY_START_PRES
    NHIST_stop = FILLNC;		! HISTORY_STOP_PRES
    NHIST_prev = FILLNC;		! HISTORY_PREVIOUS_VALUE

    nhist = 0;
    if(do_pres_adj)then;
	nhist = nhist + 1;
	do iprf = 1,N_PROF
	    if(DEBUG)then; ! DBLEV==37)then;
		write(msg,'("HIST prep hist rec ",i2," iprf: ",i2," for PRES_ADJ")')nhist,iprf;
		call wrtdbg(sub,msg);
	    endif
	    NHIST_step = "ARCA";			! HISTORY_STEP     for pres_adj
	    NHIST_soft = "ADJP";			! HISTORY_SOFTWARE for pres_adj
	    NHIST_actn = "  IP";			! HISTORY_ACTION   for pres_adj
	    NHIST_qctest = "                ";		! HISTORY_QCTEST   for pres_adj - NOT FILLED for adjustments
	    call put_ncprof_histrec(sub,ncid,nhist,iprf)
	enddo
    endif
    !-------------------------------------------------------------------------------------------
    if(do_psal_adj)then;
	nhist = nhist + 1;
	do iprf = 1,N_PROF
	    if(DEBUG)then; ! DBLEV==37)then;
		write(msg,'("HIST prep hist rec ",i2," iprf: ",i2," for PSAL_ADJ")')nhist,iprf;
		call wrtdbg(sub,msg);
	    endif
	    NHIST_step = "ARCA";			! HISTORY_STEP     for psal adj
	    NHIST_soft = "ADJS";			! HISTORY_SOFTWARE for psal adj
	    NHIST_actn = "  IP";			! HISTORY_ACTION   for psal adj
	    NHIST_qctest = "                ";		! HISTORY_QCTEST   for psal adj - NOT FILLED for adjustments
	    call put_ncprof_histrec(sub,ncid,nhist,iprf)
	enddo
    endif
    !-------------------------------------------------------------------------------------------
    nhist = nhist + 1;
    ! Until I reset the counters (some day), get qctest for ALL profiles
    do ipf = 1,N_PROF
	call qctest2hex(sub,ipf,perf_tmp,fail_tmp);
	qcperf(ipf) = perf_tmp;
	qcfail(ipf) = fail_tmp;
	if(DBLEV==31)then;		! LEV31
	    write(msg,'("QCTEST put_calhis, iprf: ",i3," perf_tmp: ",a16," fail_tmp: ",a16)')ipf,perf_tmp,fail_tmp	! LEV31
	    call wrtdbg(sub,msg);	! LEV31
	endif				! LEV31
    enddo

    do iprf = 1,N_PROF
	if(DEBUG)then; ! DBLEV==37)then;
	    write(msg,'("HIST prep hist rec ",i2," iprf: ",i2," for QC PERFORMED")')nhist,iprf;
	    call wrtdbg(sub,msg);
	endif
	NHIST_step = "ARGQ"; NHIST_soft = "QCPL"; NHIST_actn = "QCP$"; NHIST_qctest = qcperf(iprf);
	call put_ncprof_histrec(sub,ncid,nhist,iprf)
    enddo
    nhist = nhist + 1;
    do iprf = 1,N_PROF
	if(DBLEV==37)then;
	    write(msg,'("HIST prep hist rec ",i2," iprf: ",i2," for QC FAILED")')nhist,iprf;
	   call wrtdbg(sub,msg);
	endif
	NHIST_step = "ARGQ"; NHIST_soft = "QCPL"; NHIST_actn = "QCF$"; NHIST_qctest = qcfail(iprf);
	call put_ncprof_histrec(sub,ncid,nhist,iprf)
    enddo
 
    if(DBLEV==37)then; write(msg,'("LEV37 Leave sub ",a)')sub; call wrtdbg(sub,msg); endif

end subroutine put_ncprof_calhis
!================================================================================
subroutine put_ncprof_histrec(caller,ncid,ihist,iprof)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id
    integer,      intent( in ) :: ihist		! history record to write
    integer,      intent( in ) :: iprof		! profile to write

    integer :: iret
    integer :: ret = 0;
    integer ::  loc
    integer :: st2(2), ct2(2), st3(3), ct3(3), st4(4), ct4(4); ! start/count for vara uts
    character(len=18), parameter :: sub = 'put_ncprof_histrec'

    ! st2/ct2 are start/count for the three REAL fields to be put
    st2(1) = iprof; ct2(1) = 1;		
    st2(2) = ihist; ct2(2) = 1;		
    ! st3/ct3 are start/count for the char fields - the count of the first dimensin is the size of the string
    st3(1) = 1; !varies by string length
    st3(2) = iprof; ct3(2) = 1;		
    st3(3) = ihist; ct3(3) = 1;		
    ! NOTE NF_UNLIMITED This is the call that increments the UNLIMITED dimension for prof
    ct3(1) =  4;  ret=nf_put_vara_text(ncid,HIST_ids( 1),st3,ct3,NHIST_inst);   call cke(sub,ret,2101);
    ct3(1) =  4;  ret=nf_put_vara_text(ncid,HIST_ids( 2),st3,ct3,NHIST_step);   call cke(sub,ret,2102);
    ct3(1) =  4;  ret=nf_put_vara_text(ncid,HIST_ids( 3),st3,ct3,NHIST_soft);   call cke(sub,ret,2103);
    ct3(1) =  4;  ret=nf_put_vara_text(ncid,HIST_ids( 4),st3,ct3,NHIST_sfrl);   call cke(sub,ret,2104);
    ct3(1) = 64;  ret=nf_put_vara_text(ncid,HIST_ids( 5),st3,ct3,NHIST_refr);   call cke(sub,ret,2105);
    ct3(1) = 14;  ret=nf_put_vara_text(ncid,HIST_ids( 6),st3,ct3,NHIST_date);   call cke(sub,ret,2106);
    ct3(1) =  4;  ret=nf_put_vara_text(ncid,HIST_ids( 7),st3,ct3,NHIST_actn);   call cke(sub,ret,2107);
    ct3(1) = 16;  ret=nf_put_vara_text(ncid,HIST_ids( 8),st3,ct3,NHIST_parm);   call cke(sub,ret,2108);
                  ret=nf_put_vara_real(ncid,HIST_ids( 9),st2,ct2,NHIST_strt);   call cke(sub,ret,2109);
                  ret=nf_put_vara_real(ncid,HIST_ids(10),st2,ct2,NHIST_stop);   call cke(sub,ret,2110);
                  ret=nf_put_vara_real(ncid,HIST_ids(11),st2,ct2,NHIST_prev);   call cke(sub,ret,2111);
    ct3(1) = 16;  ret=nf_put_vara_text(ncid,HIST_ids(12),st3,ct3,NHIST_qctest); call cke(sub,ret,2112);

    if(DBLEV==37)then; write(msg,'("LEV37 Leave sub ",a)')sub; call wrtdbg(sub,msg); endif

end subroutine put_ncprof_histrec


!================================================================================
end module ncprof_arrays
!================================================================================
