! ===========================================================================================================
! integer function qc06_gross_test.f90
!
! -----------------------------------------------------------------------------------------------------------
!
! Revision history
! date     who    what
! 20170712 all    Release of qcplus v1.1 - many changes to all, some documented/tracked, some not
! 20160511 all    Initial release of qcplus v1.1.0
!
! ===========================================================================================================
! Notes
!
! This test applies a gross filter on observed values for pressure, temperature,
! salinity and dissolved oxygen. It needs to accommodate all of the expected
! extremes encountered in the oceans.
!
! * Pressure cannot be less than -5 dbar
! * Temperature in range -2.5 to 40.0°C
! * Salinity in range 2 to 41.0 PSU
! * Dissolved oxygen in range -0.5 to 600 micromol/kg
!
! Action: If a value fails, it should be flagged as bad data and only that value 
! need be removed from distribution on the GTS. If temperature and salinity 
! values at the same depth both fail, both values should be flagged as bad data 
! and values for depth, temperature and salinity should be removed from the TESAC 
! distributed on the GTS.
!
! Function return values:
!  0 = PASS - Success, all positions passed
!  1 = FAIL - Failure, some or all positions failed
! -1 = NOTD - cannot be determined
!

integer function qc06_gross_test(caller,info)

use communication                       ! module
use data_arrays                         ! module

implicit none

! ===========================================================================================================
! varialbe/array declarations

character(*), intent( in ) :: caller
integer,      intent( in ) :: info(3)

character(len=4), parameter :: sub = 'QC06'
character(len=21) ,parameter :: QCTEST='Gross Range Test'
integer :: result;

integer :: ifl, itype, beg, end;
integer :: ngood,nbad,nskip
real, parameter :: range_min(4) = (/   -5.0, -2.5,  2.0,  -0.5 /)
real, parameter :: range_max(4) = (/20000.0, 40.0, 41.0, 600.0 /) ! P max range just a ridiculous number

! ===========================================================================================================
! executable statements

ifl = gross_ncol;

result = NOTD ! cannot be determined

itype=info(1); beg=info(2); end=info(3); ! run characteristics

if(DBGIO) call dbg_enter_type(sub,parmname(itype))

do idat=beg,end											! idat=beg,end
    if(hasDATA(idat))then									! if hasDATA
	if(proftot(idat) > 0)then								! if ntot>0
	    ngood = 0; nbad = 0; nskip = 0;
	    do ilev=1,nlev
		if(skip_test(profflag(idat,ilev)))then
		    if(DBLEV== 6) call report_ngnbns(sub,idat,ilev,ifl,ngood,nbad,nskip+1,"skp")
		    testflag(idat,ilev,ifl) = NOTF; nskip = nskip + 1;  go to 99;
		endif ! Skip this level
		if(profdata(idat,ilev) <range_min(itype))then
		    testflag(idat,ilev,ifl) = BAD; nbad = nbad + 1;
		    if(DBLEV== 6) call report_ngnbns(sub,idat,ilev,ifl,ngood,nbad,nskip,"bad min")
		elseif(profdata(idat,ilev) > range_max(itype))then
		    testflag(idat,ilev,ifl) = BAD; nbad = nbad + 1;
		    if(DBLEV== 6) call report_ngnbns(sub,idat,ilev,ifl,ngood,nbad,nskip,"bad max")
		else
		    testflag(idat,ilev,ifl) = GOOD; ngood = ngood + 1;
		endif
		99 continue ! skip_test
	    enddo ! ilev
	endif ! if(proftot(idat) > 0									! if ntot>0
	call evaluate_flags(sub,QCTEST,datname(idat),idat,ifl,beg,result); maxresult=max(maxresult,result);
    endif	! if hasDATA
enddo		! idat

! The result returned is the max flag of those evaluated
qc06_gross_test = maxresult;
if(DBGIO) call dbg_leave_type(sub,parmname(itype),result);

return
end function qc06_gross_test
! ===========================================================================================================
